/**
 * @fileOverview  The model class Book with attribute definitions and storage management methods
 * @author Gerd Wagner
 * @author Juan-Francisco Reyes
 * @copyright Copyright � 2020-2022 Gerd Wagner (Chair of Internet Technology) and Juan-Francisco Reyes,
 * Brandenburg University of Technology, Germany.
 * @license This code is licensed under The Code Project Open License (CPOL), implying that the code is provided "as-is",
 * can be modified to create derivative works, can be redistributed, and can be used in commercial applications.
 */
import { fsDb } from "../initFirebase.mjs";
import { collection as fsColl, deleteDoc, doc as fsDoc, getDoc, getDocs, orderBy, query as fsQuery,
  setDoc, updateDoc, deleteField }
  from "https://www.gstatic.com/firebasejs/9.8.1/firebase-firestore-lite.js";
import { isNonEmptyString, nextYear, isIntegerOrIntegerString }
  from "../../lib/util.mjs";
import { NoConstraintViolation, MandatoryValueConstraintViolation, RangeConstraintViolation,
  IntervalConstraintViolation, PatternConstraintViolation, UniquenessConstraintViolation }
  from "../../lib/errorTypes.mjs";

/**
 * Constructor function for the class Book
 * @constructor
 * @param {isbn: string, title: string, year: number, edition: number?} slots - Object creation slots.
 */
class Book {
  // record parameter with the ES6 syntax for function parameter destructuring
  constructor ({isbn, title, year, edition}) {
    // assign properties by invoking implicit setters
    this.isbn = isbn;
    this.title = title;
    this.year = year;
    // edition is an optional property
    if (edition) this.edition = edition;
  };
  // all basic constraints of the isbn attribute
  get isbn() {
    return this._isbn;
  };
  // all basic constraints of the isbn attribute
  static checkIsbn( isbn) {
    if (!isbn) return new NoConstraintViolation();
    else if (typeof( isbn) !== "string" || isbn.trim() === "") {
      return new RangeConstraintViolation("The ISBN must be a non-empty string!");
    } else if (!(/\b\d{9}(\d|X)\b/.test( isbn))) {
      return new PatternConstraintViolation(
        'The ISBN must be a 10-digit string or a 9-digit string followed by "X"!');
    } else {
      return new NoConstraintViolation();
    }
  };
  // mandatory value and uniqueness constraints
  static async checkIsbnAsId( isbn) {
    let validationResult = Book.checkIsbn( isbn);
    if ((validationResult instanceof NoConstraintViolation)) {
      if (!isbn) {
        validationResult = new MandatoryValueConstraintViolation(
          "A value for the ISBN must be provided!");
      } else {
        const bookDocSn = await getDoc( fsDoc( fsDb, "books", isbn));
        if (bookDocSn.exists()) {
          validationResult = new UniquenessConstraintViolation(
             "There is already a book record with this ISBN!");
        } else {
          validationResult = new NoConstraintViolation();
        }
      }
    }
    return validationResult;
  };
  set isbn( n) {
    const validationResult = Book.checkIsbn( n);
    if (validationResult instanceof NoConstraintViolation) {
      this._isbn = n;
    } else {
      throw validationResult;
    }
  };
  get title() {
    return this._title;
  };
  static checkTitle( t) {
    if (!t) {
      return new MandatoryValueConstraintViolation("A title must be provided!");
    } else if (!isNonEmptyString( t)) {
      return new RangeConstraintViolation("The title must be a non-empty string!");
    } else {
      return new NoConstraintViolation();
    }
  };
  set title( t) {
    const validationResult = Book.checkTitle( t);
    if (validationResult instanceof NoConstraintViolation) {
      this._title = t;
    } else {
      throw validationResult;
    }
  };
  get year() {
    return this._year;
  };
  static checkYear(y) {
    const YEAR_FIRST_BOOK = 1459;
    if (!y) {
      return new MandatoryValueConstraintViolation(
        "A publication year must be provided!");
    } else if (!isIntegerOrIntegerString( y)) {
      return new RangeConstraintViolation("The value of year must be an integer!");
    } else {
      if (typeof y === "string") y = parseInt( y);
      if (y < YEAR_FIRST_BOOK || y > nextYear()) {
        return new IntervalConstraintViolation(
          `The value of year must be between ${YEAR_FIRST_BOOK} and next year!`);
      } else {
        return new NoConstraintViolation();
      }
    }
  };
  set year( y) {
    const validationResult = Book.checkYear( y);
    if (validationResult instanceof NoConstraintViolation) {
      this._year = y;
    } else {
      throw validationResult;
    }
  };
  get edition() {
    return this._edition;
  };
  static checkEdition( e) {
    // the "edition" attribute is optional
    if (!e) return new NoConstraintViolation();
    else {
      if (!isIntegerOrIntegerString( e) || parseInt( e) < 1) {
        return new RangeConstraintViolation(
          "The value of edition must be a positive integer!");
      } else {
        return new NoConstraintViolation();
      }
    }
  };
  set edition( e) {
    const validationResult = Book.checkEdition( e);
    if (validationResult instanceof NoConstraintViolation) {
      if (!e) delete this._edition;  // unset optional property
      else this._edition = e;
    } else {
      throw validationResult;
    }
  };
}
/*********************************************************
 ***  Class-level ("static") storage management methods **
 *********************************************************/
/**
 * Conversion between a Book object and a corresponding Firestore document
 * @type {{toFirestore: (function(*): {year: number,
 * isbn: (Document.isbn|*), title}), fromFirestore: (function(*, *=): Book)}}
 */
Book.converter = {
  toFirestore: function (book) {
    const data = {
      isbn: book.isbn,
      title: book.title,
      year: parseInt( book.year)
    };
    if (book.edition) data.edition = parseInt( book.edition);
    return data;
  },
  fromFirestore: function (snapshot, options) {
    const data = snapshot.data( options);
    return new Book( data);
  },
};
/**
 * Create a Firestore document in the Firestore collection "books"
 * @param slots: {object}
 * @returns {Promise<void>}
 */
Book.add = async function (slots) {
  let book = null;
  try {
    // validate data by creating Book instance
    book = new Book( slots);
    // invoke asynchronous ID/uniqueness check
    let validationResult = await Book.checkIsbnAsId( book.isbn);
    if (!validationResult instanceof NoConstraintViolation) throw validationResult;
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
    book = null;
  }
  if (book) {
    try {
      const bookDocRef = fsDoc( fsDb, "books", book.isbn).withConverter( Book.converter);
      await setDoc( bookDocRef, book);
      console.log(`Book record "${book.isbn}" created!`);
    } catch (e) {
      console.error(`${e.constructor.name}: ${e.message} + ${e}`);
    }
  }
};
/**
 * Load a book record from Firestore
 * @param isbn: {object}
 * @returns {Promise<*>} bookRecord: {array}
 */
Book.retrieve = async function( isbn) {
  try {
    const bookRec = (await getDoc( fsDoc(fsDb, "books", isbn)
      .withConverter( Book.converter))).data();
    console.log(`Book record "${bookRec.isbn}" retrieved.`);
    return bookRec;
  } catch (e) {
    console.error(`Error retrieving book record: ${e}`);
  }
};
/**
 * Load all book records from Firestore
 * @param order: {string}
 * @returns {Promise<*>} bookRecords: {array}
 */
Book.retrieveAll = async function (order) {
  if (!order) order = "isbn";
  const booksCollRef = fsColl( fsDb, "books"),
    q = fsQuery( booksCollRef, orderBy( order));
  try {
    const bookRecs = (await getDocs( q.withConverter( Book.converter))).docs.map( d => d.data());
    console.log(`${bookRecs.length} book records retrieved ${order ? "ordered by " + order : ""}`);
    return bookRecs;
  } catch (e) {
    console.error(`Error retrieving book records: ${e}`);
  }
};
/**
 * Update a Firestore document in the Firestore collection "books"
 * @param slots: {object}
 * @returns {Promise<void>}
 */
Book.update = async function (slots) {
  let noConstraintViolated = true,
    validationResult = null,
    bookBeforeUpdate = null;
  const bookDocRef = fsDoc( fsDb, "books", slots.isbn).withConverter( Book.converter),
    updatedSlots = {};
  try {
    // retrieve up-to-date book record
    const bookDocSn = await getDoc( bookDocRef);
    bookBeforeUpdate = bookDocSn.data();
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  try {
    if (bookBeforeUpdate.title !== slots.title) {
      validationResult = Book.checkTitle( slots.title);
      if (validationResult instanceof NoConstraintViolation) updatedSlots.title = slots.title;
      else throw validationResult;
    }
    if (bookBeforeUpdate.year !== parseInt( slots.year)) {
      validationResult = Book.checkYear( slots.year);
      if (validationResult instanceof NoConstraintViolation) updatedSlots.year = parseInt( slots.year);
      else throw validationResult;
    }
    if (slots.edition && bookBeforeUpdate.edition !== parseInt( slots.edition)) {
      // slots.edition has a non-empty value that is different from the old value
      validationResult = Book.checkEdition( slots.edition);
      if (validationResult instanceof NoConstraintViolation) updatedSlots.edition = parseInt( slots.edition);
      else throw validationResult;
    } else if (!slots.edition && bookBeforeUpdate.edition) {
      // slots.edition has an empty value while the old value was not empty
      updatedSlots.edition = await updateDoc( bookDocRef, {edition: deleteField()});
    }
  } catch (e) {
    noConstraintViolated = false;
    console.error(`${e.constructor.name}: ${e.message}`);
  }
  if (noConstraintViolated) {
    const updatedProperties = Object.keys(updatedSlots);
    if (updatedProperties.length) {
      await updateDoc(bookDocRef, updatedSlots);
      console.log(`Property(ies) "${updatedProperties.toString()}" modified for book record "${slots.isbn}"`);
    } else {
      console.log(`No property value changed for book record "${slots.isbn}"!`);
    }
  }
};
/**
 * Delete a Firestore document from the Firestore collection "books"
 * @param isbn: {string}
 * @returns {Promise<void>}
 */
Book.destroy = async function (isbn) {
  try {
    await deleteDoc( fsDoc(fsDb, "books", isbn));
    console.log(`Book record "${isbn}" deleted!`);
  } catch (e) {
    console.error(`Error deleting book record: ${e}`);
  }
};
/*******************************************
 *** Auxiliary methods for testing **********
 ********************************************/
/**
 * Create test data
 */
Book.generateTestData = async function() {
  try {
    console.log("Generating test data...");
    const response = await fetch( "../../test-data/books.json");
    const bookRecs = await response.json();
    await Promise.all( bookRecs.map( d => Book.add( d)));
    console.log(`${bookRecs.length} books saved.`);
  } catch (e) {
    console.error(`${e.constructor.name}: ${e.message}`);
  }
};
/**
 * Clear database
 */
Book.clearData = async function() {
  if (confirm("Do you really want to delete all book records?")) {
    try {
      console.log("Clearing test data...");
      const booksCollRef = fsColl( fsDb, "books");
      const booksQrySn = (await getDocs( booksCollRef));
      await Promise.all( booksQrySn.docs.map( d => Book.destroy( d.id)))
      console.log(`${booksQrySn.docs.length} books deleted.`);
    } catch (e) {
      console.error(`${e.constructor.name}: ${e.message}`);
    }
  }
};

export default Book;
