/**
 * @fileOverview  View methods for the use case "update book"
 * @author Gerd Wagner
 * @author Juan-Francisco Reyes
 */
/***************************************************************
 Import classes, data types and utility procedures
 ***************************************************************/
import { handleAuthentication } from "./accessControl.mjs";
import Book from "../m/Book.mjs";
import { fillSelectWithOptions } from "../../lib/util.mjs";

/***************************************************************
 Setup and handle UI Authentication
 ***************************************************************/
handleAuthentication();

/***************************************************************
 Load data
 ***************************************************************/
const bookRecords = await Book.retrieveAll();

/***************************************************************
 Declare variables for accessing UI elements
 ***************************************************************/
const formEl = document.forms["Book"],
  updateButton = formEl["commit"],
  selectBookEl = formEl["selectBook"];

/***************************************************************
 Add event listeners for responsive validation
 ***************************************************************/
formEl["title"].addEventListener("input", function () {
  formEl["title"].setCustomValidity(
    Book.checkTitle( formEl["title"].value).message);
});
formEl["year"].addEventListener("input", function () {
  formEl["year"].setCustomValidity(
    Book.checkYear( formEl["year"].value).message);
});
formEl["edition"].addEventListener("input", function () {
  formEl["edition"].setCustomValidity(
    Book.checkEdition( formEl["edition"].value).message);
});

/***************************************************************
 Set up (choice) widgets
 ***************************************************************/
fillSelectWithOptions(bookRecords, selectBookEl, "isbn", "title");
// when a book is selected, populate the form with its data
selectBookEl.addEventListener("change", async function () {
  const bookKey = selectBookEl.value;
  if (bookKey) {
    // retrieve up-to-date book record
    const bookRecord = await Book.retrieve( bookKey);
    for (const field of ["isbn", "title", "year", "edition"]) {
      formEl[field].value = bookRecord[field] !== undefined ? bookRecord[field] : "";
      // delete custom validation error message which may have been set before
      formEl[field].setCustomValidity("");
    }
  } else {
    formEl.reset();
  }
});

/********************************************************************
 Add further event listeners, especially for the update/submit button
 ********************************************************************/
updateButton.addEventListener("click", function () {
  const formEl = document.forms["Book"],
    selectBookEl = formEl["selectBook"],
    bookIdRef = selectBookEl.value;
  if (!bookIdRef) return;
  const slots = {
    isbn: formEl["isbn"].value,
    title: formEl["title"].value,
    year: formEl["year"].value
  };
  // set error messages in case of constraint violations
  formEl["title"].addEventListener("input", function () {
    formEl["title"].setCustomValidity(
      Book.checkTitle( slots.title).message);
  });
  formEl["year"].addEventListener("input", function () {
    formEl["year"].setCustomValidity(
      Book.checkYear( slots.year).message);
  });
  if (formEl["edition"].value) {
    formEl["edition"].addEventListener("input", function () {
      formEl["edition"].setCustomValidity(
        Book.checkEdition(slots.edition).message);
    });
  }
  if (formEl.checkValidity()) {
    Book.update( slots);
    // update the selection list option
    selectBookEl.options[selectBookEl.selectedIndex].text = slots.title;
    formEl.reset();
  }
});
// neutralize the submit event
formEl.addEventListener("submit", function (e) {
  e.preventDefault();
});
